package frc.robot.subsystems;

import java.util.function.DoubleSupplier;

import com.ctre.phoenix.motorcontrol.can.WPI_VictorSPX;

import edu.wpi.first.math.controller.ElevatorFeedforward;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.util.sendable.SendableBuilder;
import edu.wpi.first.wpilibj.Encoder;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.RepeatCommand;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import edu.wpi.first.wpilibj2.command.WaitCommand;
import frc.robot.constants.ElevatorConstants;
import frc.robot.constants.ElevatorConstants.ElevatorPositions;

public class Elevator extends SubsystemBase {
    private WPI_VictorSPX motor1;
    private WPI_VictorSPX motor2;

    private Encoder encoder;

    private PIDController positionalController;
    private PIDController velocityController;

    private ElevatorFeedforward eff;

    public Elevator() {
        motor1 = new WPI_VictorSPX(ElevatorConstants.kMotor1CANID);
        motor2 = new WPI_VictorSPX(ElevatorConstants.kMotor2CANID);

        motor2.follow(motor1);

        encoder = new Encoder(ElevatorConstants.kEncoderChannelADIO, ElevatorConstants.kEncoderChannelBDIO);
        encoder.setDistancePerPulse(ElevatorConstants.kEncoderDistancePerPulse);

        positionalController = new PIDController(
            ElevatorConstants.kPPositional, 
            ElevatorConstants.kIPositional, 
            ElevatorConstants.kDPositional
        );

        velocityController = new PIDController(
            ElevatorConstants.kPVelocity, 
            ElevatorConstants.kIVelocity, 
            ElevatorConstants.kDVelocity
        );

        eff = new ElevatorFeedforward(
            ElevatorConstants.kFFS, 
            ElevatorConstants.kFFG, 
            ElevatorConstants.kFFV
        );
    }

    public double getEncoderDistance() {
        return encoder.getDistance();
    }

    public double getEncoderRate() {
        return encoder.getRate();
    }

    public Command goToPosition(ElevatorPositions position, double timeout) {
        return run(() -> {
            double voltsOut = positionalController.calculate(
                encoder.getDistance(),
                position.getPositionValue()
            ) + eff.calculate(0);

            motor1.setVoltage(voltsOut);
        }).until(positionalController::atSetpoint).withTimeout(timeout);
    }

    public Command assistedManualControl(DoubleSupplier speed) {
        return run(() -> {
            double realSpeedTarget = speed.getAsDouble() * ElevatorConstants.kElevatorMaxVelocity;

            double voltsOut = velocityController.calculate(
                encoder.getRate(), 
                realSpeedTarget
            ) + eff.calculate(realSpeedTarget);

            motor1.setVoltage(voltsOut);
        });
    }

    public Command debugSetpoints() {
        return new RepeatCommand(
            runOnce(() -> {
                System.out.println("Current Bottom Positional Setpoint: " + ElevatorPositions.kBottom.getPositionValue());
                System.out.println("Current Low-Mid Positional Setpoint: " + ElevatorPositions.kLowMid.getPositionValue());
                System.out.println("Current High-Mid Positional Setpoint: " + ElevatorPositions.kHighMid.getPositionValue());
                System.out.println("Current Top Positional Setpoint: " + ElevatorPositions.kTop.getPositionValue());
                System.out.println("");
            }).andThen(new WaitCommand(5))
        );
    }

    @Override
    public void initSendable(SendableBuilder builder) {
        builder.setActuator(true);

        builder.addDoubleProperty(
            ".currentBottomPositionalSetpoint", 
            () -> ElevatorPositions.kBottom.getPositionValue(), 
            (d) -> ElevatorPositions.kBottom.setPositionValue(d)
        );

        builder.addDoubleProperty(
            ".currentLowMidPositionalSetpoint", 
            () -> ElevatorPositions.kLowMid.getPositionValue(), 
            (d) -> ElevatorPositions.kLowMid.setPositionValue(d)
        );

        builder.addDoubleProperty(
            ".currentHighMidPositionalSetpoint", 
            () -> ElevatorPositions.kHighMid.getPositionValue(), 
            (d) -> ElevatorPositions.kHighMid.setPositionValue(d)
        );

        builder.addDoubleProperty(
            ".currentTopPositionalSetpoint", 
            () -> ElevatorPositions.kTop.getPositionValue(), 
            (d) -> ElevatorPositions.kTop.setPositionValue(d)
        );
    }
}
